#include <bdd-for-c.h>
#include "../../common/IZ_common.h"
#include "../../../../__mocks__/src/packages/stdinc/IZ_stdlib.mock.h"
#include "IZ_list.h"

bool NodeExists(IZ_ListNode** node, u64 _index, IZ_List* list) {
	return *((u64*) (*node)->value) == 42069;
}

bool NodeExists2(IZ_ListNode** node, u64 _index, IZ_List* list) {
	return *((u64*) (*node)->value) == 69420;
}

bool NodeDoesNotExist(IZ_ListNode** node, u64 _index, IZ_List* list) {
	return *((u64*) (*node)->value) == 55555;
}

spec("data") {
	describe("list") {
		describe("Initialize") {
			static IZ_List list;

			after_each() {
				IZ_ListTeardown(&list);
			}

			it("sets root to NULL") {
				IZ_ListInitialize(&list);

				check(
					list.root == NULL,
					"List not properly initialized. Root: %p", list.root
				);
			}
		}

		describe("Teardown") {
			static IZ_List list;

			before_each() {
				static u64 value1 = 69420u;
				static u64 value2 = 42069u;
				static u64 value3 = 69069u;

				IZ_ListInitialize(&list);
				IZ_ListAppendNode(&list, &value1);
				IZ_ListAppendNode(&list, &value2);
				IZ_ListAppendNode(&list, &value3);
			}

			after_each() {
				mock_reset(IZ_free);
			}

			it("removes all nodes from the list") {
				mock_set_expected_calls(IZ_free, 4);
				IZ_ListTeardown(&list);
				unsigned int expected_calls = mock_get_expected_calls(IZ_free);
				unsigned int actual_calls = mock_get_actual_calls(IZ_free);

				check(
					expected_calls == actual_calls,
					"Deallocator function call count mismatch. Expected %u, got %u.", expected_calls, actual_calls
				);
			}
		}

		describe("AppendNode") {
			static IZ_List list;
			static IZ_List list2;

			before_each() {
				IZ_ListInitialize(&list);
			}

			before_each() {
				IZ_ListInitialize(&list2);
				static u64 existing_value = 69420u;
				static IZ_ListNode existing_node = {
					.value = &existing_value,
					.next = NULL,
				};
				list2.root = &existing_node;
				list2.length = 1;
			}

			after_each() {
				mock_reset(IZ_malloc);
			}

			after_each() {
				IZ_ListTeardown(&list);
			}

			after_each() {
				IZ_ListTeardown(&list2);
			}

			it("appends new node to empty list and sets it as root") {
				static u64 value = 69420u;
				IZ_ListAppendNode(&list, &value);
				u64 added_value = *((u64*) list.root->value);

				check(added_value == 69420u, "Node not properly appended. Value: %u", added_value);
				check(
					mock_is_called(IZ_malloc),
					"Allocator function not called."
				);
				check(list.length == 1, "Length mismatch.");
			}

			it("appends new node to non-empty list") {
				static u64 value1 = 42069u;

				IZ_ListAppendNode(&list2, &value1);

				check(*((u64*) list2.root->next->value) == 42069u, "Node not properly appended.");
				check(
					mock_is_called(IZ_malloc),
					"Allocator function not called."
				);
				check(list2.length == 2, "Length mismatch. Length: %u", list2.length);
			}
		}

		describe("FindFirstNode") {
			static IZ_List list;
			static u64 value1 = 69420u;
			static u64 value2 = 42069u;

			before_each() {
				IZ_ListInitialize(&list);
				IZ_ListAppendNode(&list, &value1);
				IZ_ListAppendNode(&list, &value2);
			}

			after_each() {
				IZ_ListTeardown(&list);
			}

			it("retrieves first node satisfying the filter condition") {
				static IZ_ListNode** node;
				node = IZ_ListFindFirstNode(&list, NodeExists);
				check(*((u64*) (*node)->value) == 42069u, "Existing node not found.");
			}

			it("returns NULL when all nodes do not satisfy the filter condition") {
				static IZ_ListNode** node;
				node = IZ_ListFindFirstNode(&list, NodeDoesNotExist);
				check(node == NULL, "Non-existing node found.");
			}
		}

		describe("DeleteNode") {
			static IZ_List list;

			before_each() {
				static u64 value1 = 69420u;
				static u64 value2 = 42069u;
				static u64 value3 = 69069u;

				IZ_ListInitialize(&list);
				IZ_ListAppendNode(&list, &value1);
				IZ_ListAppendNode(&list, &value2);
				IZ_ListAppendNode(&list, &value3);
			}

			after_each() {
				mock_reset(IZ_free);
			}

			it("removes first node satisfying the filter condition") {
				IZ_ListNode** existing_node = IZ_ListFindFirstNode(&list, NodeExists);
				IZ_ListDeleteNode(&list, existing_node);

				check(
					mock_is_called(IZ_free),
					"Deallocator function not called."
				);

				check(
					IZ_ListFindFirstNode(&list, NodeExists2),
					"Node supposed to be present in list is absent."
				);

				check(
					!IZ_ListFindFirstNode(&list, NodeExists),
					"Deleted node still present in list."
				);

				check(list.length == 2, "Length mismatch.");
			}
		}
	}
}
